#include "ScannerSAnalysisUtils.h"

using namespace std;

double TableData::Min(int i){
  if(i<MinRanges.size() || i>-1)
    return MinRanges[i];
  else{
    std::clog <<"Error in: double TableData::Min(int i) -- Out of bounds i="<<i<<endl;
    exit(EXIT_FAILURE);
  }
}

double TableData::Max(int i){
  if(i<MaxRanges.size() || i>-1)
    return MaxRanges[i];
  else{
    std::clog <<"Error in: double TableData::Max(int i) -- Out of bounds i="<<i<<endl;
    exit(EXIT_FAILURE);
  }
}

int TableData::Ngrid(int i){
  if(i<NsGrid.size() || i>-1)
    return NsGrid[i];
  else{
    std::clog <<"Error in: double TableData::Ngrid(int i) -- Out of bounds i="<<i<<endl;
    exit(EXIT_FAILURE);
  }
}

//This functions clears all objects that hold the data
void TableData::clear_data(void){
  grid.clear(); //clear grid of points
  MinRanges.clear(); //clear ranges
  MaxRanges.clear(); //..
  unit_cell_sizes.clear(); //clear sizes of unitary cell edges
  NsGrid.clear(); //clear vector of number of points along each direction 
  NprodCoeff.clear(); //clear vector of coefficients in the index expansion
  func=NULL; //function to be interpolated pointing to invalid address
}

void TableData::calcNprodCoeffs(){ //Calculate the coefficients of the expansion that takes the tensor product indices into the linear grid table index
  NprodCoeff[0]=1;
  for(vector<int>::size_type i=1;i<NprodCoeff.size();++i){
    NprodCoeff[i]=NprodCoeff[i-1]*NsGrid[i-1];
  }
}

int TableData::calc_grid_coord(std::vector<int> & indices){ //Calculate the actual coordinate on the grid
  int coord=0;
  for(vector<int>::size_type i=0;i<indices.size();++i)
    coord+=NprodCoeff[i]*indices[i];
  return coord;
}

//algorithm to compute the D-dimensional coordinates (indices) of a point stored in the linear grid  at position A.
void TableData::calcpartition(int A,int N,std::vector<int> & indices){
  //This algorithm solves for A=a0+N0(a1+N1(a2+N2(...(a_{n-2}+N_{n-2}a_{n-1})))) using integer division
  int q=A,qnext,n=NsGrid.size();
  for(int i=0;i<n-1;++i){
    qnext=q/NsGrid[i];
    indices[i]=q-NsGrid[i]*qnext;
    q=qnext;
  }
  indices[n-1]=q;
}

//Same algorithm as above but for a unit cell hypercube
void TableData::calcpartition_unit(int A,int N,std::vector<int> & indices){
  int q=A,qnext,n=NsGrid.size();
  for(int i=0;i<n-1;++i){
    qnext=q/2;
    indices[i]=q-2*qnext;
    q=qnext;
  }
  indices[n-1]=q;
}

void TableData::create_from_scratch(std::vector<double> & MinRanges_in,std::vector<double> & MaxRanges_in,std::vector<int> & NsGrid_in,double (*func_in)(std::vector<double> & Point,void *),void * parfunc_in){
  clear_data(); //Clear class data to fill up tables from scratch
  //Write new input into class
  MinRanges=MinRanges_in; 
  MaxRanges=MaxRanges_in;
  NsGrid=NsGrid_in;
  unit_cell_sizes.resize(NsGrid.size());
  NprodCoeff.resize(NsGrid.size());
  calcNprodCoeffs();
  func=func_in;
  parfunc=parfunc_in;
  //Compute the dimensions of the grid vector to hold the data
  int Ngrid=1;
  for(vector<int>::size_type i=0;i<NsGrid.size();++i){
    unit_cell_sizes[i]=(MaxRanges[i]-MinRanges[i])/(NsGrid[i]-1);//Also compute the sizes of the edges of the unit cell
    Ngrid*=NsGrid[i];
  }
  vector<int> indices(NsGrid.size()); //vector to hold index coordinates on the grid
  vector<double> Point(NsGrid.size());  //vector to hold actual coordinates of the point
  //Cmputing values of the function on the grid vector
  for(int i=0;i<Ngrid;++i){
    calcpartition(i,Ngrid,indices); //Find grid indices of the grid point i
    for(vector<double>::size_type n=0;n<Point.size();++n){
      Point[n]=MinRanges[n]+indices[n]*unit_cell_sizes[n]; //Compute coordinates of the corresponding point
    }
    grid.push_back(func(Point,parfunc)); //Compute function value at the point and store on the grid vector
  }

}

void TableData::create_from_scratch_flush(std::vector<double> & MinRanges_in,std::vector<double> & MaxRanges_in,std::vector<int> & NsGrid_in,double (*func_in)(std::vector<double> & Point,void *),void * parfunc_in,const string fileName){

  ofstream os(fileName.c_str());
  
  clear_data(); //Clear class data to fill up tables from scratch
  //Write new input into class
  MinRanges=MinRanges_in; 
  MaxRanges=MaxRanges_in;
  NsGrid=NsGrid_in;
  unit_cell_sizes.resize(NsGrid.size());
  NprodCoeff.resize(NsGrid.size());
  calcNprodCoeffs();
  func=func_in;
  parfunc=parfunc_in;
  //Compute the dimensions of the grid vector to hold the data
  int Ngrid=1;
  for(vector<int>::size_type i=0;i<NsGrid.size();++i){
    unit_cell_sizes[i]=(MaxRanges[i]-MinRanges[i])/(NsGrid[i]-1);//Also compute the sizes of the edges of the unit cell
    Ngrid*=NsGrid[i];
  }

  os<<NsGrid.size()<<endl; //Output the dimension of the grid
  for(vector<int>::size_type i=0;i<NsGrid.size();++i)
    os<<NsGrid[i]<<"\t"<<MinRanges[i]<<"\t"<<MaxRanges[i]<<endl; //output a line with ranges and number of points for each direction

  vector<int> indices(NsGrid.size()); //vector to hold index coordinates on the grid
  vector<double> Point(NsGrid.size());  //vector to hold actual coordinates of the point
  //Cmputing values of the function on the grid vector
  for(int i=0;i<Ngrid;++i){
    calcpartition(i,Ngrid,indices); //Find grid indices of the grid point i
    for(vector<double>::size_type n=0;n<Point.size();++n){
      Point[n]=MinRanges[n]+indices[n]*unit_cell_sizes[n]; //Compute coordinates of the corresponding point
    }
    grid.push_back(func(Point,parfunc)); //Compute function value at the point and store on the grid vector
    os<<grid.back()<<"\t"<<flush; //output the grid vector as it is calculated
  }
  os<<endl; // end file.
  os.close();
}

//Interpolation function
double TableData::operator()(std::vector<double> & Point){
  //compute cell coordinates and shifts
  vector<int> cell; //vector to hold grid coordinates of the reference vertex on the interpolation cell
  vector<double> dcell,Point0; // coordinates of the interpolation point inside the unit cell (fractional displacements) and coordinates of reference vertex point.
  for(vector<double>::size_type i=0;i<Point.size();++i){
    if(Point[i]<MinRanges[i] ||Point[i]>MaxRanges[i]){
      cerr<<"Error in double TableData::operator()(std::vector<double> & Point): Point out of table bounds."<<endl<<"ABORTING! Please fix your code."<< endl;
      exit(EXIT_FAILURE);
    }
  }

  if(Point[Point.size()-1]==MaxRanges[Point.size()-1]) {
    Point[Point.size()-1]-=1e-6*unit_cell_sizes[Point.size()-1];
  }
  for(vector<int>::size_type i=0;i<NsGrid.size();++i){
    cell.push_back((Point[i]-MinRanges[i])/unit_cell_sizes[i]); // Compute coordinates of reference vertex
    dcell.push_back((Point[i]-MinRanges[i])/unit_cell_sizes[i]-cell.back()); // Compute fractional displacements
    Point0.push_back(MinRanges[i]+cell.back()*unit_cell_sizes[i]); //Compute coordinates of reference vertex
  }

  //Loop over vertex of hypercubic cell
  int n=pow(2,NsGrid.size()); //number of vertices for an  NsGrid.size() dimensional hypercube.
  vector<int> indices(NsGrid.size()); //variable to hold binary coordinates of the vertex on the unit-hypercubic cell
  ////////////vector<double> PointVertex(NsGrid.size()); //variable to hold actual coordinates of the vertex
  int igrid0=calc_grid_coord(cell);
  double out=0;
  for(int i=0;i<n;++i){
    calcpartition_unit(i,n,indices); //Compute binary coordinates of the vertex
    double weight=1; //interpolation weigth of the current vertex
    for(vector<int>::size_type k=0;k<indices.size();++k){
      weight*=abs(dcell[k]-(1-indices[k])); //computing fractional volume between interpolation point and (diagonally) opposite vertex.
      /////////////// PointVertex[k]=Point0[k]+indices[k]*unit_cell_sizes[k]; //Coordinates of actual vertex
    }
    /////////////!!!!!!!!!!!!!!!!!!!!!!!!!!!!!??????????????????
    //out+=weight*func(PointVertex);//Add contributiuon from this vertex to output value
    //std::clog<<"????? ="<<igrid0<< endl;
    out+=weight*grid[igrid0+calc_grid_coord(indices)];//Add contributiuon from this vertex to output value
  }

  return out; //return the interpolated estimate
}

//Overloaded insertion operator
std::ostream & operator<<(std::ostream & os, const TableData & T){
  os<<T.NsGrid.size()<<endl; //Output the dimension of the grid
  for(vector<int>::size_type i=0;i<T.NsGrid.size();++i)
    os<<T.NsGrid[i]<<"\t"<<T.MinRanges[i]<<"\t"<<T.MaxRanges[i]<<endl; //output a line with ranges and number of points for each direction
  for(vector<double>::size_type i=0;i<T.grid.size();++i)
    os<<T.grid[i]<<"\t"; //output the grid vector
  os<<endl;
  return os;
}

//Overloaded extraction operator
std::istream & operator>>(std::istream & is, TableData & T){
  T.clear_data(); //Clear data to fill up tables from scratch
  int dim; 
  is>>dim; //read dimension of grid
  //Re-size vector accordingly
  T.NsGrid.resize(dim); 
  T.MinRanges.resize(dim);
  T.MaxRanges.resize(dim);
  T.unit_cell_sizes.resize(dim);  
  T.NprodCoeff.resize(dim);

  int dim_grid=1;//  read number of points, ranges  and compute  grid vector size simultaneously
  for(vector<int>::size_type i=0;i<dim;++i){
    is>>T.NsGrid[i]>>T.MinRanges[i]>>T.MaxRanges[i];
    T.unit_cell_sizes[i]=(T.MaxRanges[i]-T.MinRanges[i])/(T.NsGrid[i]-1);//Also compute the sizes of the edges of the unit cell
    dim_grid*=T.NsGrid[i];
  }

  T.calcNprodCoeffs();
  //Fill up grid vector with values read from file
  T.grid.resize(dim_grid);
  for(vector<double>::size_type i=0;i<dim_grid;++i)
    is>>T.grid[i];
  return is;
}

void TableData::save_table(const char *fileName){
  bool filetest=fileExists(fileName); //Check if file exists

  if(filetest){ //If yes, ask user what to do
    cerr<<"WARNING: File \""<<fileName<<"\" exists. Do you want to over-write it? [Y/N]"<< endl;
    char answer;
    cin>>answer;
    if(answer!='y' && answer!='Y'){
      filetest=false;
      cerr<<"RUN ABORTED. Check what you really want to do with \""<<fileName<<"\"."<< endl;
      exit(EXIT_FAILURE);
    }
  }

  ofstream out(fileName); //Open output file
  out<<*this; //Write to output file
  out.close(); //Close file
  
}

// Check a file already exists on disk
bool fileExists(const char *fileName)
{
    ifstream infile(fileName);
    return infile.good();
}

void TableData::create_from_file(const char *fileName){
  bool filetest=fileExists(fileName); //Check if file exists

  if(filetest){ 
    ifstream infile(fileName);
    infile>>*this;
    infile.close();
  }
  else{
    cerr<<"RUN ABORTED. File \""<<fileName<<"\" does not seem to exist."<< endl;
    exit(EXIT_FAILURE);
  }
  
}

// Least square fits
double fit_multi_params(double * Xij_in, double * y_in, int npoints, double *coeffs, int ncoeffs, double * err){

  double chisq;
  gsl_matrix *X, *cov;
  gsl_vector *y, *c;

  X = gsl_matrix_alloc(npoints,ncoeffs);
  y = gsl_vector_alloc(npoints);

  c = gsl_vector_alloc (ncoeffs);
  cov = gsl_matrix_alloc (ncoeffs,ncoeffs);

  for(int i=0; i<npoints;++i){
    for(int j= 0; j<ncoeffs;++j)
      gsl_matrix_set (X,i,j,Xij_in[i*ncoeffs+j]);
    gsl_vector_set (y,i,y_in[i]);
  }
  
  {
    gsl_multifit_linear_workspace * work = gsl_multifit_linear_alloc (npoints,ncoeffs);
    gsl_multifit_linear(X,y,c,cov,&chisq, work);
    gsl_multifit_linear_free (work);
  }
  
  for(int i=0;i<ncoeffs;++i){
    coeffs[i]=gsl_vector_get(c,i);
    err[i]=sqrt(gsl_matrix_get(cov,i,i));
  }

  gsl_matrix_free (X);
  gsl_vector_free (y);
  gsl_vector_free (c);
  gsl_matrix_free (cov);
  return 1-chisq/gsl_stats_tss(y_in,1,npoints);
}


#ifdef HBMODE
//signal strength check
bool CheckSignalStrengths_gen(const int nsigma, const double sep, const HBobject & HBA,std::map<string,double> & results){
  using exp_c::m_h;
  using exp_c::BRsm_gamgam;
  using exp_c::BRsm_ZZ;
  using exp_c::BRsm_WW;
  using exp_c::BRsm_tautau;
  using exp_c::BRsm_bb;

  const double exp_mu_VoF_min = exp_c::mu_VoF - nsigma * exp_c::mu_VoF_m;
  const double exp_mu_VoF_max = exp_c::mu_VoF + nsigma * exp_c::mu_VoF_p;

  const double exp_mu_gamgam_min = exp_c::mu_gamgam - nsigma * exp_c::mu_gamgam_m;
  const double exp_mu_gamgam_max = exp_c::mu_gamgam + nsigma * exp_c::mu_gamgam_p;

  const double exp_mu_ZZ_min = exp_c::mu_ZZ - nsigma * exp_c::mu_ZZ_m;
  const double exp_mu_ZZ_max = exp_c::mu_ZZ + nsigma * exp_c::mu_ZZ_p;

  const double exp_mu_WW_min = exp_c::mu_WW - nsigma * exp_c::mu_WW_m;
  const double exp_mu_WW_max = exp_c::mu_WW + nsigma * exp_c::mu_WW_p;

  const double exp_mu_tautau_min = exp_c::mu_tautau - nsigma * exp_c::mu_tautau_m;
  const double exp_mu_tautau_max = exp_c::mu_tautau + nsigma * exp_c::mu_tautau_p;

  const double exp_mu_bb_min = exp_c::mu_bb - nsigma * exp_c::mu_bb_m;
  const double exp_mu_bb_max = exp_c::mu_bb + nsigma * exp_c::mu_bb_p;

  //create array of particles on peak
  std::vector<int> on_peak;
  for(int i=0;i!=HBA.nHzero;++i)
    if(abs(HBA.Mh[i]-m_h)<sep)
      on_peak.push_back(i);

  //return false if multible or no neutral particles are on the peak
  if(on_peak.size()!=1)
    return false;

  //return false if charged particles are on the peak
  for(int i=0;i!=HBA.nHplus;++i)
    if(abs(HBA.MHplus[i]-m_h)<sep)
      return false;

  //initialize results
  results={
    {"mu_V",0},
    {"mu_F",0},
    {"mu_VoF",0},
    {"mu_gamgam",0},
    {"mu_ZZ",0},
    {"mu_WW",0},
    {"mu_tautau",0},
    {"mu_bb",0}};

  //calculate signal strengths
  for(int i : on_peak){
    results["mu_V"]     += HBA.CS_lhc8_hjW_ratio[i];
    results["mu_F"]     += HBA.CS_lhc8_hj_ratio[i];
    results["mu_gamgam"]+= HBA.CS_lhc8_hj_ratio[i]*HBA.BR_hjgaga[i]/BRsm_gamgam;
    results["mu_ZZ"]    += HBA.CS_lhc8_hj_ratio[i]*HBA.BR_hjZZ[i]/BRsm_ZZ;
    results["mu_WW"]    += HBA.CS_lhc8_hj_ratio[i]*HBA.BR_hjWW[i]/BRsm_WW;
    results["mu_tautau"]+= HBA.CS_lhc8_hj_ratio[i]*HBA.BR_hjtautau[i]/BRsm_tautau;
    results["mu_bb"]    += HBA.CS_lhc8_hj_ratio[i]*HBA.BR_hjbb[i]/BRsm_bb;
  }
  results["mu_VoF"]=results["mu_V"]/results["mu_F"];

  //check experimental bounds
  if( (results["mu_VoF"]>exp_mu_VoF_min)      && (results["mu_VoF"]<exp_mu_VoF_max)
   && (results["mu_gamgam"]>exp_mu_gamgam_min)&& (results["mu_gamgam"]<exp_mu_gamgam_max)
   && (results["mu_ZZ"]>exp_mu_ZZ_min)        && (results["mu_ZZ"]<exp_mu_ZZ_max)
   && (results["mu_WW"]>exp_mu_WW_min)        && (results["mu_WW"]<exp_mu_WW_max)
   && (results["mu_tautau"]>exp_mu_tautau_min)&& (results["mu_tautau"]<exp_mu_tautau_max)
   && (results["mu_bb"]>exp_mu_bb_min)        && (results["mu_bb"]<exp_mu_bb_max))
    return true;

  return false;
}
#endif
