#include "ScannerSC2HDM.h"
#include <iomanip>
using namespace std;

string C2HDM::filepath="ScannerScore/ScannerSModels/C2HDM/C2HDM.potential";
const double C2HDM::Ecm2 = 2000.; //todo: make these global
const double C2HDM::Ecm7 = 7000.;
const double C2HDM::Ecm8 = 8000.;
int C2HDM::fail_global=0;
int C2HDM::fail_STU=0;
int C2HDM::fail_Z_bbar=0;
int C2HDM::fail_bs_gamma=0;
int C2HDM::fail_HB=0;
int C2HDM::fail_rates=0;
int C2HDM::fail_EDM=0;


C2HDM::C2HDM(){
//CONSTRUCTOR

#ifdef HBMODE
  HiggsBoundsAnalysis.rewrite_values(nHzero,nHplus,whichanalyses,whichinput); //sets the values needed to initialize HiggsBounds (will be initialized in UserInitCalcs)
#endif

  //Loading numerical tables for cross-sections
  Load_ggF_bbF_noEW_SusHiPP_PPbarTables_LHC_plus_Tevatron();
  //Load numerical Superiso tables
  TableSuperisoBsgammaType1.create_from_file("ScannerScore/data/SuperisoBsgamma2HDMType1.dat");
  TableSuperisoBsgammaType2.create_from_file("ScannerScore/data/SuperisoBsgamma2HDMType2.dat");
  TableSuperisoBsgammaType3.create_from_file("ScannerScore/data/SuperisoBsgamma2HDMType3.dat");
  TableSuperisoBsgammaType4.create_from_file("ScannerScore/data/SuperisoBsgamma2HDMType4.dat");

  //Initialise arrays to zero

  //Higgs couplings to vector bosons
  Rhvv.fill(0);
  for(auto& r : Rhhv)
    r.fill(0);
  //3x3 Mixing Matrix
  for(auto& r : Rmix)
    r.fill(0);
}

void C2HDM::readfile(std::istream & in){
//This function is called from the overloaded operator >> when a C2HDM input file is provided.
//On the base class, Potential::readfile simply calls openinput() on "in".
//since C2HDM files only contain the relevant variables, "this" readfile opens an auxiliary file in the same format as Potential::readfile
//and then replaces the relevante variables bty the values in "in".
  ifstream modelfile;
  OpenModelFile(modelfile,filepath);
  openinput(modelfile);

  mass_bounds[1]=getvariable(in,Cp.Mass.vmin(1),Cp.Mass.vmax(1),">m2");
  mass_bounds[2]=getvariable(in,Cp.Mass.vmin(2),Cp.Mass.vmax(2),">m3");
  mass_bounds[3]=getvariable(in,Cp.Mass.vmin(3),Cp.Mass.vmax(3),">mH+");
  mass_bounds[4]=getvariable(in,Cp.Mass.vmin(4),Cp.Mass.vmax(4),">mH+");
  L_bounds[8]=getvariable(in,Cp.L.vmin(8),Cp.L.vmax(8),">re_m12^2");

  m1_reparam=getvariable(in,MLpars.vmin(0),MLpars.vmax(0),">m2(mH+)");
  bool tanbeta_bounds=getvariable(in,PhiPar.vmin(0),PhiPar.vmax(0),">tanbeta");

  getvariable(in,Type2HDM,">2HDM_type:");

  //check of all independent parameters were defined in the input file
  for(int i=2;i!=5;++i)
    if(!mass_bounds[i]){
      cerr<<"ScannerS ERROR: Unable to find a necessary parameter in the C2HDM input file. This model ALWAYS requires you define the range for m3, mH+, tanbeta, re_m12^2 and either m2 or m2(mH+). These are independent parameters that will be randomly generated."<<endl;
      exit(EXIT_FAILURE);
    }
  if((!L_bounds[8]) || (!tanbeta_bounds)){
    cerr<<"ScannerS ERROR: Unable to find a necessary parameter in the C2HDM input file. This model ALWAYS requires you define the range for m3, mH+, tanbeta, re_m12^2 and either m2 or m2(mH+). These are independent parameters that will be randomly generated."<<endl;
    exit(EXIT_FAILURE);
  }
  if(!(mass_bounds[1]||m1_reparam)){
    cerr<<"ScannerS ERROR: Unable to find a necessary parameter in the C2HDM input file. This model ALWAYS requires you define the range for m3, mH+, tanbeta, re_m12^2 and either m2 or m2(mH+). These are independent parameters that will be randomly generated."<<endl;
    exit(EXIT_FAILURE);
  }

  if(mass_bounds[1] && m1_reparam)
    clog<<"ScannerS Warning in C2HDM::readfile(): The input file contained bounds on both m2 and m2(mH+). m2(mH+) will be used!"<<endl;

  if(!m1_reparam)
    MLpars=MLparamVec();//do not use reparametrization if no ranges are given (resets MLpars to empty)

  //check if there were bounds imposed on the dependent couplings
  L_bounds[0]=getvariable(in,Cp.L.vmin(0),Cp.L.vmax(0),">m11^2");
  L_bounds[1]=getvariable(in,Cp.L.vmin(1),Cp.L.vmax(1),">m22^2");
  L_bounds[2]=getvariable(in,Cp.L.vmin(2),Cp.L.vmax(2),">lambda1");
  L_bounds[3]=getvariable(in,Cp.L.vmin(3),Cp.L.vmax(3),">lambda2");
  L_bounds[4]=getvariable(in,Cp.L.vmin(4),Cp.L.vmax(4),">lambda3");
  L_bounds[5]=getvariable(in,Cp.L.vmin(5),Cp.L.vmax(5),">lambda4");
  L_bounds[6]=getvariable(in,Cp.L.vmin(6),Cp.L.vmax(6),">re_lambda5");
  L_bounds[7]=getvariable(in,Cp.L.vmin(7),Cp.L.vmax(7),">im_lambda5");
  L_bounds[9]=getvariable(in,Cp.L.vmin(9),Cp.L.vmax(9),">im_m12^2");
  mass_bounds[0]=getvariable(in,Cp.Mass.vmin(0),Cp.Mass.vmax(0),">m1");
  //check if the pseudoscalar admixture for the input states is bounded
  pseudo_bounds[0]=getvariable(in,pseudo_min[0],pseudo_max[0],">pseudo1");
  pseudo_bounds[1]=getvariable(in,pseudo_min[1],pseudo_max[1],">pseudo2");
  pseudo_bounds[2]=getvariable(in,pseudo_min[2],pseudo_max[2],">pseudo3");

}

std::string C2HDM::name() const{
//returns the name of the type of potential used.
  string name("C2HDM");
  return name;
}

void C2HDM::PrintHeader() const {

  std::clog<<endl<<"***********************************************"<<endl;
  std::clog<<"*********** "<<scanners_version<<" ************"<<endl;
  std::clog<<"***********************************************" <<endl;
  std::clog<<endl<<"---------- SUMMARY INFO FOR THIS RUN ---------- "<<endl;
  std::clog<<endl<<"Potential Type: C2HDM"<<endl;
  std::clog<<endl<<"V(Phi1,Phi2) = L0Phi1'Phi1 + L1Phi2'Phi2 - (L8+iL9)Phi1'Phi2 - (L8-iL9)Phi2'Phi1 + (L2/2)(Phi1'Phi1)² + (L3/2)(Phi2'Phi2)² + L4Phi1'Phi1Phi2'Phi2 + L5Phi2'Phi1Phi1'Phi2 + 1/2 ((L6+iL7)(Phi1'Phi2)²+(L6-iL7)(Phi2'Phi1)²)"<<endl;
  std::clog<<endl<< "The expanded potential is:"<<endl;
  std::clog<<endl<<(*this)<<endl;
  std::clog<<endl<<"The ranges of parameters to be scanned are:"<<endl;
  std::clog<<endl<<"Ranges for vacuum expectation values:"<<endl;
  for(size_t i=0;i!=Phi.size();++i){
    if(phi_bounds[i])
      std::clog<<"phi"<<i<<" in ["<<Phi.vgetmin(i)<<":"<<Phi.vgetmax(i)<<"]"<< endl;
  }
  std::clog<<endl<<"Ranges for couplings:"<< endl;
  for(size_t i=0;i!=Cp.L.size();++i){
    if(L_bounds[i])
      std::clog<<"L"<<i<<" in ["<<Cp.L.vgetmin(i)<<":"<<Cp.L.vgetmax(i)<<"]"<< endl;
  }
  std::clog<<endl<<"Ranges for physical state masses:"<< endl;
  for(size_t i=0;i!=Cp.Mass.size();++i){
    if(mass_bounds[i])
    std::clog<<"m"<<i<<" in ["<<Cp.Mass.vgetmin(i)<<":"<<Cp.Mass.vgetmax(i)<<"]"<<endl;
  }
  if(m1_reparam)
    clog<<"m1 in [m3"<<MLpars.vgetmin(0)<<":m3+"<<MLpars.vgetmax(0)<<"]"<< endl;
  std::clog<<endl<<"Range for tanbeta (v2/v6 parameterization):"<<endl;
  std::clog<<"tanbeta in ["<<PhiPar.vgetmin(0)<<":"<<PhiPar.vgetmax(0)<<"]"<<endl;
  std::clog<<endl;

  std::clog<<endl<<"Allowed Cp odd admixture to the input mass states:"<< endl;
  for(size_t i=0;i!=pseudo_bounds.size();++i){
    if(pseudo_bounds[i])
    std::clog<<"pseudoscalar admixture of state "<<i<<" in ["<<pseudo_min[i]<<":"<<pseudo_max[i]<<"]"<<endl;
  }

  std::clog<<"The seed you are using for the random number generator is: "<<r.getseed()<< endl<< endl;



}

bool C2HDM::CheckBounds() const{
//unlike Potential::CheckBounds(), this only checks dependent parameters for which the range was explicity declared in the 2HDM input file. if they were not declared in input file (and were loaded of 2HDM.potential by default) L_bounds will be false and the parameters will not be checked

  for(size_t i=0;i!=Cp.L.size();++i)//L_bounds size checked at PrintHeader
    if(L_bounds[i] && (Cp.L[i]<Cp.L.vgetmin(i) || Cp.L[i]>Cp.L.vgetmax(i))){
#ifdef VERBOSE
    std::clog<<endl<<"---> ScannerS: Point rejected by C2HDM::CheckBounds()."<<endl;
#endif
      return false;
  }

  if(mass_bounds[0] && (Cp.Mass[0]<Cp.Mass.vgetmin(0) || Cp.Mass[0]>Cp.Mass.vgetmax(0))){
#ifdef VERBOSE
  std::clog<<endl<<"---> ScannerS: Point rejected by C2HDM::CheckBounds()."<<endl;
#endif
    return false;
}


double beta=atan(Phi[6]/Phi[2]);
double sb = sin(beta);
double cb = cos(beta);
for(size_t i=0;i!=pseudo_bounds.size();++i)
  if(pseudo_bounds[i] && ( pow(-sb*Mmixing[i][3]+cb*Mmixing[i][7],2)*100 < pseudo_min[i] || pow(-sb*Mmixing[i][3]+cb*Mmixing[i][7],2)*100 > pseudo_max[i])){
#ifdef VERBOSE
  std::clog<<endl<<"---> ScannerS: Point rejected by N2HDM::CheckBounds()."<<endl;
#endif
    return false;
}

  // sanity check of the found mixing
  // checks mass degeneracy of H+ and H-
  if(fabs(Cp.Mass[3]-Cp.Mass[4])>1e-3){
    clog<<"Warning in C2HDM::CheckBounds: illegal mixing input recieved, point will be dropped!"<<endl;
    return false;
  }


  return true;
}

#ifdef HBMODE
void C2HDM::PrepareHiggsBoundsInput(){

  //Call of Hdecay to compute all branching ratios
  CreateInputFileHdecayC2HDM(Type2HDM,tanbeta,a1,a2,a3, mH1, mH2,mHc,Cp.L[8]);
  HdecayCalcC2HDM(Hdecay);


  /////////////////////////////////////////////////////////
  //Passing couplings, masses and decay properties to HB //
  /////////////////////////////////////////////////////////
  //Light Higgs - H1
  HiggsBoundsAnalysis.Mh[0]=mH1;
  HiggsBoundsAnalysis.CP[0]=0;
  //Branching ratios and total width
  HiggsBoundsAnalysis.BR_hjss[0] = Hdecay.at("BR(H1 -> s sbar)");
  HiggsBoundsAnalysis.BR_hjcc[0] = Hdecay.at("BR(H1 -> c cbar)");
  HiggsBoundsAnalysis.BR_hjbb[0] = Hdecay.at("BR(H1 -> b bbar)");
  HiggsBoundsAnalysis.BR_hjmumu[0] = Hdecay.at("BR(H1 -> mu+ mu-)");
  HiggsBoundsAnalysis.BR_hjtautau[0] = Hdecay.at("BR(H1 -> tau+ tau-)");
  HiggsBoundsAnalysis.BR_hjWW[0] = Hdecay.at("BR(H1 -> W+ W-)");
  HiggsBoundsAnalysis.BR_hjZZ[0] = Hdecay.at("BR(H1 -> Z Z)");
  HiggsBoundsAnalysis.BR_hjZga[0]= Hdecay.at("BR(H1 -> Z gamma)");
  HiggsBoundsAnalysis.BR_hjgaga[0] = Hdecay.at("BR(H1 -> gamma gamma)");
  HiggsBoundsAnalysis.BR_hjgg[0] = Hdecay.at("BR(H1 -> g g)");
  HiggsBoundsAnalysis.MhGammaTot[0] = Hdecay.at("Width(H1)");

  //Medium Higgs - H2
  HiggsBoundsAnalysis.Mh[1]=mH2;
  HiggsBoundsAnalysis.CP[1]=0;
  //Branching ratios and total width
  HiggsBoundsAnalysis.BR_hjss[1] = Hdecay.at("BR(H2 -> s sbar)");
  HiggsBoundsAnalysis.BR_hjcc[1] = Hdecay.at("BR(H2 -> c cbar)");
  HiggsBoundsAnalysis.BR_hjbb[1] = Hdecay.at("BR(H2 -> b bbar)");
  HiggsBoundsAnalysis.BR_hjmumu[1] = Hdecay.at("BR(H2 -> mu+ mu-)");
  HiggsBoundsAnalysis.BR_hjtautau[1] = Hdecay.at("BR(H2 -> tau+ tau-)");
  HiggsBoundsAnalysis.BR_hjWW[1] = Hdecay.at("BR(H2 -> W+ W-)");
  HiggsBoundsAnalysis.BR_hjZZ[1] = Hdecay.at("BR(H2 -> Z Z)");
  HiggsBoundsAnalysis.BR_hjZga[1]= Hdecay.at("BR(H2 -> Z gamma)");
  HiggsBoundsAnalysis.BR_hjgaga[1] = Hdecay.at("BR(H2 -> gamma gamma)");
  HiggsBoundsAnalysis.BR_hjgg[1] = Hdecay.at("BR(H2 -> g g)");
  HiggsBoundsAnalysis.BR_hjhihi[1][0] = Hdecay.at("BR(H2 -> H1 H1)");
  HiggsBoundsAnalysis.MhGammaTot[1] = Hdecay.at("Width(H2)");
 //
  //Heavy Higgs - H3
  HiggsBoundsAnalysis.Mh[2]=mH3;
  HiggsBoundsAnalysis.CP[2]=0;
  //Branching ratios and total width
  HiggsBoundsAnalysis.BR_hjss[2] = Hdecay.at("BR(H3 -> s sbar)");
  HiggsBoundsAnalysis.BR_hjcc[2] = Hdecay.at("BR(H3 -> c cbar)");
  HiggsBoundsAnalysis.BR_hjbb[2] = Hdecay.at("BR(H3 -> b bbar)");
  HiggsBoundsAnalysis.BR_hjmumu[2] = Hdecay.at("BR(H3 -> mu+ mu-)");
  HiggsBoundsAnalysis.BR_hjtautau[2] = Hdecay.at("BR(H3 -> tau+ tau-)");
  HiggsBoundsAnalysis.BR_hjWW[2] = Hdecay.at("BR(H3 -> W+ W-)");
  HiggsBoundsAnalysis.BR_hjZZ[2] = Hdecay.at("BR(H3 -> Z Z)");
  HiggsBoundsAnalysis.BR_hjZga[2]= Hdecay.at("BR(H3 -> Z gamma)");
  HiggsBoundsAnalysis.BR_hjgaga[2] = Hdecay.at("BR(H3 -> gamma gamma)");
  HiggsBoundsAnalysis.BR_hjgg[2] = Hdecay.at("BR(H3 -> g g)");
  HiggsBoundsAnalysis.BR_hjhihi[2][0] = Hdecay.at("BR(H3 -> H1 H1)");
  HiggsBoundsAnalysis.BR_hjhihi[2][1] = Hdecay.at("BR(H3 -> H2 H2)");
  HiggsBoundsAnalysis.MhGammaTot[2] = Hdecay.at("Width(H3)");


  //Charged Higgs - H^+
  HiggsBoundsAnalysis.MHplus[0] = mHc;
  HiggsBoundsAnalysis.BR_tWpb[0] = Hdecay.at("BR(t -> b W+)");
  HiggsBoundsAnalysis.BR_tHpjb[0] = Hdecay.at("BR(t -> b H+)");
  HiggsBoundsAnalysis.BR_Hpjcs[0] = Hdecay.at("BR(H+ -> c sbar)");
  HiggsBoundsAnalysis.BR_Hpjcb[0] = Hdecay.at("BR(H+ -> c bbar)");
  HiggsBoundsAnalysis.BR_Hptaunu[0] = Hdecay.at("BR(H+ -> tau+ nu_tau)");
  HiggsBoundsAnalysis.MHplusGammaTot[0] = Hdecay.at("Width(H+)");

  //Neutral Higgs hadronic input
  for(size_t i=0;i!=nHzero;++i){
    //LEP
    HiggsBoundsAnalysis.CS_lep_hjZ_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_lep_bbhj_ratio[i]=pow(Rhf[i][1][0],2)+pow(Rhf[i][1][1],2);
    HiggsBoundsAnalysis.CS_lep_tautauhj_ratio[i]=pow(Rhf[i][2][0],2)+pow(Rhf[i][2][1],2);
    for (size_t j=0; j!=nHzero;++j)
      HiggsBoundsAnalysis.CS_lep_hjhi_ratio[i][j]=pow(Rhhv[i][j],2);
    //Tevatron
    HiggsBoundsAnalysis.CS_tev_hj_ratio[i]=(InternalTableSusHiPPbar_ggF_noEW(Ecm2,HiggsBoundsAnalysis.Mh[i],1,Rhf[i][0][0],Rhf[i][1][0])
      +InternalTableSusHiPPbar_ggF_noEW(Ecm2,HiggsBoundsAnalysis.Mh[i],-1,Rhf[i][0][1],Rhf[i][1][1])
      +InternalTableSusHiPPbar_bbF_noEW(Ecm2,HiggsBoundsAnalysis.Mh[i],1,Rhf[i][1][0])
      +InternalTableSusHiPPbar_bbF_noEW(Ecm2,HiggsBoundsAnalysis.Mh[i],-1,Rhf[i][1][1]))/InternalTableSusHiPPbar_ggF_noEW(Ecm2,HiggsBoundsAnalysis.Mh[i],1,1,1);
    HiggsBoundsAnalysis.CS_tev_hjb_ratio[i]=pow(Rhf[i][1][0],2)+pow(Rhf[i][1][1],2);
    HiggsBoundsAnalysis.CS_tev_hjW_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_tev_hjZ_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_tev_vbf_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_tev_tthj_ratio[i]=pow(Rhf[i][0][0],2)+pow(Rhf[i][0][1],2);
    //LHC7
    HiggsBoundsAnalysis.CS_lhc7_hj_ratio[i]=(InternalTableSusHiPP_ggF_noEW(Ecm7,HiggsBoundsAnalysis.Mh[i],1,Rhf[i][0][0],Rhf[i][1][0])
      +InternalTableSusHiPP_ggF_noEW(Ecm7,HiggsBoundsAnalysis.Mh[i],-1,Rhf[i][0][1],Rhf[i][1][1])
      +InternalTableSusHiPP_bbF_noEW(Ecm7,HiggsBoundsAnalysis.Mh[i],1,Rhf[i][1][0])
      +InternalTableSusHiPP_bbF_noEW(Ecm7,HiggsBoundsAnalysis.Mh[i],-1,Rhf[i][1][1]))/InternalTableSusHiPP_ggF_noEW(Ecm7,HiggsBoundsAnalysis.Mh[i],1,1,1);
    HiggsBoundsAnalysis.CS_lhc7_hjb_ratio[i]=pow(Rhf[i][1][0],2)+pow(Rhf[i][1][1],2);
    HiggsBoundsAnalysis.CS_lhc7_hjW_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_lhc7_hjZ_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_lhc7_vbf_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_lhc7_tthj_ratio[i]=pow(Rhf[i][0][0],2)+pow(Rhf[i][0][1],2);
    //LHC8
    HiggsBoundsAnalysis.CS_lhc8_hj_ratio[i]=(InternalTableSusHiPP_ggF_noEW(Ecm8,HiggsBoundsAnalysis.Mh[i],1,Rhf[i][0][0],Rhf[i][1][0])
      +InternalTableSusHiPP_ggF_noEW(Ecm8,HiggsBoundsAnalysis.Mh[i],-1,Rhf[i][0][1],Rhf[i][1][1])
      +InternalTableSusHiPP_bbF_noEW(Ecm8,HiggsBoundsAnalysis.Mh[i],1,Rhf[i][1][0])
      +InternalTableSusHiPP_bbF_noEW(Ecm8,HiggsBoundsAnalysis.Mh[i],-1,Rhf[i][1][1]))/InternalTableSusHiPP_ggF_noEW(Ecm8,HiggsBoundsAnalysis.Mh[i],1,1,1);
    HiggsBoundsAnalysis.CS_lhc8_hjb_ratio[i]=pow(Rhf[i][1][0],2)+pow(Rhf[i][1][1],2);
    HiggsBoundsAnalysis.CS_lhc8_hjW_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_lhc8_hjZ_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_lhc8_vbf_ratio[i]=pow(Rhvv[i],2);
    HiggsBoundsAnalysis.CS_lhc8_tthj_ratio[i]=pow(Rhf[i][0][0],2)+pow(Rhf[i][0][1],2);
  }

 //Charged Higgs hadronic input
  HiggsBoundsAnalysis.CS_lep_HpjHmi_ratio[0] = 1;
}

bool C2HDM::CheckSignalStrengths() {
  //check at two sigma
  constexpr int nsigma = 2;
  //with a seperation of 5GeV
  constexpr double seperation = 5.;

  return CheckSignalStrengths_gen(nsigma,seperation,HiggsBoundsAnalysis,mu_results);
}

#endif

bool C2HDM::TestZbbarbLEP() const{
  // if (Type2HDM==1 || Type2HDM==3)return true;
  return tanbeta>Interp(mHc,CKMfitterZbbar);
}

bool C2HDM::Test_bs_gamma(){
  //flavour Physics observables
  const double bsgamma_max=4.1;
  const double bsgamma_min=3.;
  vector<double>massbeta {mHc,beta};
  if(Type2HDM==1)
    bsgamma=TableSuperisoBsgammaType1(massbeta);  ////Call superiso table to test B-> X_s gamma
  else if(Type2HDM==2){
    if(mHc < 480){
      bsgamma=0;
      return false; //Misiak et al. 2015
    }
    bsgamma=TableSuperisoBsgammaType2(massbeta);  ////Call superiso table to test B-> X_s gamma
  }
  else if(Type2HDM==3)
    bsgamma=TableSuperisoBsgammaType3(massbeta);  ////Call superiso table to test B-> X_s gamma
  else if(Type2HDM==4){
    if(mHc < 480){
      bsgamma=0;
      return false; //Misiak et al. 2015
    }
    bsgamma=TableSuperisoBsgammaType4(massbeta);  ////Call superiso table to test B-> X_s gamma
  }
  if(bsgamma*1e4>bsgamma_max || bsgamma*1e4 <bsgamma_min)
      return false;
  return true;
}

void C2HDM::PhysicalStatesC2HDM(){
  //This routines returns the masses of the scalars in the 2HDM as well as the alpha mixing angle for the 2 CP even Higgs bosons and the beta angle between the vevs.
  //	INPUT:
  //       Mass - internal (vector like) array of masses
  //       Mixing - internal (matrix like) mixing matrix
  //       Phi - internal (vector like) array of field values
  //	OUTPUT:
  //       mHc - Mass of the charged Higgs
  //    - mass ordering
  //       mH1 - lightest CP even Higgs
  //       mH2 - medium mass CP even Higgs
  //       mH3 - heaviest CP even Higgs
  //       a1,a2,a3 mixing angles in mass ordering
  //    - singlet ordering
  //       singlet admixtures sH1,sH2,sH3


  //Computing beta from VEVs
  tanbeta=Phi[6]/Phi[2];
  beta=atan(tanbeta);

  double sb = sin(beta);
  double cb = cos(beta);
  std::array<std::array<double,3>,3> Rmix0  = {
    std::array<double,3>{Mmixing[0][2],Mmixing[0][6],-sb*Mmixing[0][3]+cb*Mmixing[0][7]},
    std::array<double,3>{Mmixing[1][2],Mmixing[1][6],-sb*Mmixing[1][3]+cb*Mmixing[1][7]},
    std::array<double,3>{Mmixing[2][2],Mmixing[2][6],-sb*Mmixing[2][3]+cb*Mmixing[2][7]},
  };

  //check of proper rotation
  for(int i = 0;i!=3;i++){
    double test = fabs(cb*Mmixing[i][3]+sb*Mmixing[i][7]);
    if(test>1e-5){
      clog<<"Warning: Goldstone rotation did not work out! ("<<test<<"!=0)"<<endl;
      break;
    }
  }



  mHc=Cp.Mass[4]; //Mass of the charged Higgs (numbering checked in VERBOSE mode. Note Mass[4] is the same because it corresponds to the imaginary part).

  std::array<unsigned int,3> h {0,1,2};
  int temp;

  if(Cp.Mass[h[0]]>Cp.Mass[h[2]]){
    temp=h[2];
    h[2]=h[0];
    h[0]=temp;
  }
  if(Cp.Mass[h[0]]>Cp.Mass[h[1]]){
    temp=h[1];
    h[1]=h[0];
    h[0]=temp;
  }
  else if(Cp.Mass[h[1]]>Cp.Mass[h[2]]){
    temp=h[2];
    h[2]=h[1];
    h[1]=temp;
  }

  mH1 = Cp.Mass[h[0]];
  mH2 = Cp.Mass[h[1]];
  mH3 = Cp.Mass[h[2]];

  double det=-Rmix0[h[0]][2]*Rmix0[h[1]][1]*Rmix0[h[2]][0] + Rmix0[h[0]][1]*Rmix0[h[1]][2]*Rmix0[h[2]][0] + Rmix0[h[0]][2]*Rmix0[h[1]][0]*Rmix0[h[2]][1] - Rmix0[h[0]][0]*Rmix0[h[1]][2]* Rmix0[h[2]][1] - Rmix0[h[0]][1]*Rmix0[h[1]][0]*Rmix0[h[2]][2] + Rmix0[h[0]][0]*Rmix0[h[1]][1]*Rmix0[h[2]][2];

  // all angles shall be constrained to [-Pi/2,+Pi/2],
  // which means that the elements containing only cos need to be positive
  // to ensure the proper determinant,
  // either an even or odd number of line sign changes (=reflect_) is needed
  std::array<int,3>reflect {1,1,1};
  if(det>0){
    //clog<<"positive determinant"<<endl;
    if(Rmix0[h[0]][0]<0 && Rmix0[h[2]][2]<0){
      reflect[0]=-1;
      reflect[2]=-1;
    }
    else if(Rmix0[h[0]][0]<0){
      reflect[0]=-1;
      reflect[1]=-1;
    }
    else if(Rmix0[h[2]][2]<0){
      reflect[1]=-1;
      reflect[2]=-1;
    }
  }
  else{
    if(Rmix0[h[0]][0]<0 && Rmix0[h[2]][2]<0){
      reflect[0]=-1;
      reflect[1]=-1;
      reflect[2]=-1;
    }
    else if(Rmix0[h[0]][0]<0){
      reflect[0]=-1;
    }
    else if(Rmix0[h[2]][2]<0){
      reflect[2]=-1;
    }
    else{//both positive
      reflect[1]=-1;
    }
  }

  //compute the reordered Rmix
  for(int i=0;i!=3;i++)
    for(int j = 0; j!=3; j++)
      Rmix[i][j]=Rmix0[h[i]][j]*reflect[i];

  //constrain a3 to be positive
  // if(Rmix[1][2]<0){
  //   Rmix[0][2]*=-1;
  //   Rmix[1][2]*=-1;
  //   Rmix[2][0]*=-1;
  //   Rmix[2][1]*=-1;
  // }

  //compute the alphas from the mixing matrix
  a2=asin(Rmix[0][2]);
  a1=asin(Rmix[0][1]/cos(a2));
  a3=asin(Rmix[1][2]/cos(a2));
  // clog<<"alphas: "<<a1<<" "<<a2<<" "<<a3<<endl;

  // calculate couplings to fermions and gauge bosons (in here because EDM check needs those)

  // Rhf[iHiggs][ifermion][CP]: iHiggs=0,1,2 (A,H1,H2), ifermion= 0,1,2 (up-type quarks, d-type quarks and leptons), CP = 0, 1 (even,odd)
  if(Type2HDM==1){	//Type I
    for(int i=0;i!=nHzero;i++){
      Rhf[i][0][0] = Rmix[i][1]/sb;
      Rhf[i][0][1] = -Rmix[i][2]/tanbeta;
      Rhf[i][1][0] = Rmix[i][1]/sb;
      Rhf[i][1][1] = Rmix[i][2]/tanbeta;
      Rhf[i][2][0] = Rmix[i][1]/sb;
      Rhf[i][2][1] = Rmix[i][2]/tanbeta;
    }
  }
  if(Type2HDM==2){	//Type II
    for(int i=0;i!=nHzero;i++){
      Rhf[i][0][0] = Rmix[i][1]/sb;
      Rhf[i][0][1] = -Rmix[i][2]/tanbeta;
      Rhf[i][1][0] = Rmix[i][0]/cb;
      Rhf[i][1][1] = -Rmix[i][2]*tanbeta;
      Rhf[i][2][0] = Rmix[i][0]/cb;
      Rhf[i][2][1] = -Rmix[i][2]*tanbeta;
    }
  }
  if(Type2HDM==3){	//Lepton Specific
    for(int i=0;i!=nHzero;i++){
      Rhf[i][0][0] = Rmix[i][1]/sb;
      Rhf[i][0][1] = -Rmix[i][2]/tanbeta;
      Rhf[i][1][0] = Rmix[i][1]/sb;
      Rhf[i][1][1] = Rmix[i][2]/tanbeta;
      Rhf[i][2][0] = Rmix[i][0]/cb;
      Rhf[i][2][1] = -Rmix[i][2]*tanbeta;
    }
  }
  if(Type2HDM==4){	//flipped
    for(int i=0;i!=nHzero;i++){
      Rhf[i][0][0] = Rmix[i][1]/sb;
      Rhf[i][0][1] = -Rmix[i][2]/tanbeta;
      Rhf[i][1][0] = Rmix[i][0]/cb;
      Rhf[i][1][1] = -Rmix[i][2]*tanbeta;
      Rhf[i][2][0] = Rmix[i][1]/sb;
      Rhf[i][2][1] = Rmix[i][2]/tanbeta;
    }
  }

  //hiVV couplings
  for(int i = 0;i!=nHzero;++i)
    Rhvv[i]=cb*Rmix[i][0]+sb*Rmix[i][1];

  //hihjZ couplings
  for(int i=0;i!=nHzero;++i){
    for(int j=0; j!=nHzero;++j)
      Rhhv[i][j]=(sb*Rmix[i][0]-cb*Rmix[i][1])*Rmix[j][2] - (sb*Rmix[j][0]-cb*Rmix[j][1])*Rmix[i][3];
  }

}


bool C2HDM::CheckTreeUni(){
  maxev = std::max({
    std::abs( 3/2. * (Cp.L[2]+Cp.L[3]) + sqrt( 9/4. * pow(Cp.L[2]-Cp.L[3],2) + pow(2*Cp.L[4]+Cp.L[5],2) ) ),
    std::abs( 3/2. * (Cp.L[2]+Cp.L[3]) - sqrt( 9/4. * pow(Cp.L[2]-Cp.L[3],2) + pow(2*Cp.L[4]+Cp.L[5],2) ) ),
    std::abs( 1/2. * (Cp.L[2]+Cp.L[3]) + 1/2. * sqrt( pow(Cp.L[2]-Cp.L[3],2) + 4*Cp.L[5]*Cp.L[5] ) ),
    std::abs( 1/2. * (Cp.L[2]+Cp.L[3]) - 1/2. * sqrt( pow(Cp.L[2]-Cp.L[3],2) + 4*Cp.L[5]*Cp.L[5] ) ),
    std::abs( 1/2. * (Cp.L[2]+Cp.L[3]) + 1/2. * sqrt( pow(Cp.L[2]-Cp.L[3],2) + 4*(Cp.L[6]*Cp.L[6]+Cp.L[7]*Cp.L[7]) ) ),
    std::abs( 1/2. * (Cp.L[2]+Cp.L[3]) - 1/2. * sqrt( pow(Cp.L[2]-Cp.L[3],2) + 4*(Cp.L[6]*Cp.L[6]+Cp.L[7]*Cp.L[7]) ) ),
    std::abs( Cp.L[4] + 2*Cp.L[5] + 3*sqrt(Cp.L[6]*Cp.L[6]+Cp.L[7]*Cp.L[7]) ),
    std::abs( Cp.L[4] + 2*Cp.L[5] - 3*sqrt(Cp.L[6]*Cp.L[6]+Cp.L[7]*Cp.L[7]) ),
    std::abs( Cp.L[4] + sqrt(Cp.L[6]*Cp.L[6]+Cp.L[7]*Cp.L[7]) ),
    std::abs( Cp.L[4] - sqrt(Cp.L[6]*Cp.L[6]+Cp.L[7]*Cp.L[7]) ),
    std::abs( Cp.L[4] + Cp.L[5] ),
    std::abs( Cp.L[4] - Cp.L[5] )
  });
  maxev /= acos(-1);
  if(maxev > 8)
    return false;
  return true;
}

bool C2HDM::TestSTU() {
  if(CheckVirtual(mH1,mHc) &&
      CheckVirtual(mH1,mH1) &&
      CheckVirtual(mHc,mHc)) {


    double cb = cos(beta);
    double sb = sin(beta);
  //

    // feed values into the STU input
    // matrices
    STU_C2HDM.m_U.at(0).at(0) = cb;
    STU_C2HDM.m_U.at(0).at(1) = -sb;
    STU_C2HDM.m_U.at(1).at(0) = sb;
    STU_C2HDM.m_U.at(1).at(1) = cb;
    STU_C2HDM.m_Vr.at(0).at(0) = 0;
    STU_C2HDM.m_Vr.at(0).at(1) = Rmix[0][0];
    STU_C2HDM.m_Vr.at(0).at(2) = Rmix[1][0];
    STU_C2HDM.m_Vr.at(0).at(3) = Rmix[2][0];
    STU_C2HDM.m_Vr.at(1).at(0) = 0;
    STU_C2HDM.m_Vr.at(1).at(1) = Rmix[0][1];
    STU_C2HDM.m_Vr.at(1).at(2) = Rmix[1][1];
    STU_C2HDM.m_Vr.at(1).at(3) = Rmix[2][1];
    STU_C2HDM.m_Vi.at(0).at(0) = cb;
    STU_C2HDM.m_Vi.at(0).at(1) = -sb*Rmix[0][2];
    STU_C2HDM.m_Vi.at(0).at(2) = -sb*Rmix[1][2];
    STU_C2HDM.m_Vi.at(0).at(3) = -sb*Rmix[2][2];
    STU_C2HDM.m_Vi.at(1).at(0) = sb;
    STU_C2HDM.m_Vi.at(1).at(1) = cb*Rmix[0][2];
    STU_C2HDM.m_Vi.at(1).at(2) = cb*Rmix[1][2];
    STU_C2HDM.m_Vi.at(1).at(3) = cb*Rmix[2][2];
    //masses
    STU_C2HDM.mHzero.at(0)=mH1;
    STU_C2HDM.mHzero.at(1)=mH2;
    STU_C2HDM.mHzero.at(2)=mH3;
    STU_C2HDM.mHplus.at(0)=mHc;

    // clog<<"U={{";
    // for(const auto & line: STU_C2HDM.m_U){
    //   for(const auto & x: line){
    //     clog<<x<<",";
    //   }
    //   clog<<"\b},{";
    // }
    // clog<<"\b"<<"\b}\n";

    // clog<<"Vr={{";
    // for(const auto & line: STU_C2HDM.m_Vr){
    //   for(const auto & x: line){
    //     clog<<x<<",";
    //   }
    //   clog<<"\b},{";
    // }
    // clog<<"\b\b\b}}\n";

    // clog<<"Vi={{";
    // for(const auto & line: STU_C2HDM.m_Vi){
    //   for(const auto & x: line){
    //     clog<<x<<",";
    //   }
    //   clog<<"\b},{";
    // }
    // clog<<"\b\b\b}}\n";


    //evaluate
    bool result = STU_C2HDM.evaluate();
    // clog<<"S "<<STU_C2HDM.S;
    // clog<<"\tT "<<STU_C2HDM.T;
    // clog<<"\tU "<<STU_C2HDM.U<<endl;
    return result;
  } else {
    return false;
    // alternative behaviour: ignore test for this point
    // std::clog<<"Warning in C2HDM::TestSTU: Attempt to aplly STU test for too low masses. Test ignored for this point."<<endl;
    // STU_C2HDM.S = 0;
    // STU_C2HDM.T = 0;
    // STU_C2HDM.U = 0;
    // return true;
  }
}

//calculate and check bounds on electron electric dipole moment
// calculation from 1311.4704, all formula numbers refer to this
bool C2HDM::TestEDM() {
  double edm =0;
  edm -= fermion_loops();//sign from B.1
  // clog<<std::setprecision(9)<<"ferm:\t"<<-fermion_loops()*exp_c::m_e *sqrt(2) * exp_c::Gf / pow(4* acos(-1),4)<<endl;
  edm += Hc_loops();//sign from B.5
  // clog<<std::setprecision(9)<<"hc:\t"<<Hc_loops()*exp_c::m_e *sqrt(2) * exp_c::Gf / pow(4* acos(-1),4)<<endl;
  edm += W_loops();//sign from B.7
  // clog<<std::setprecision(9)<<"W:\t"<<W_loops()*exp_c::m_e *sqrt(2) * exp_c::Gf / pow(4* acos(-1),4)<<endl;
  edm -= HW_loops();//sign from B.9
  // clog<<std::setprecision(9)<<"HW:\t"<<-HW_loops()*exp_c::m_e *sqrt(2) * exp_c::Gf / pow(4* acos(-1),4)<<endl;
  edm *= exp_c::m_e *sqrt(2) * exp_c::Gf / pow(4* acos(-1),4);

  eEDM = edm;
  // clog<<"electron EDM: "<<edm<<endl;
  return (fabs(edm) < 7.02e-15);//ACME 1310.7534 90% exclusion limit (converted to 1/GeV)
}

double C2HDM::fermion_loops() { // B.1
  using exp_c::f_Q;
  using exp_c::f_Nc;
  using exp_c::f_coup_gam;
  using exp_c::mf_sq;
  using exp_c::f_coup_Z;
  using exp_c::mZ_sq;

  const std::array<double,3> mH0_sq {mH1*mH1,mH2*mH2,mH3*mH3};
  double result = 0;
  for(size_t f=0; f!=3; f++) {//iterate over top, bottom, tau
    for(size_t h=0; h!=nHzero; h++){ //iterate over neutral Higgses
      // gammagamma contribution
      double temp = 0;
      temp += Rhf[h][2][1] * Rhf[h][f][0] * f_coup_gam[f] * mf_sq[f] / mH0_sq[h] * (-1) * func_I(mf_sq[f],mH0_sq[h],1);
      temp += Rhf[h][2][0] * Rhf[h][f][1] * f_coup_gam[f] * mf_sq[f] / mH0_sq[h] * (-1) * func_I(mf_sq[f],mH0_sq[h],2);
      result += f_Nc[f] * f_Q[f] * f_coup_gam[2] * temp;
      // Zgamma contribution
      temp = 0;
      temp += Rhf[h][2][1] * Rhf[h][f][0] * f_coup_Z[f] * mf_sq[f] / (mH0_sq[h] - mZ_sq) *(func_I(mf_sq[f],mZ_sq,1) - func_I(mf_sq[f],mH0_sq[h],1));
      temp += Rhf[h][2][0] * Rhf[h][f][1] * f_coup_Z[f] * mf_sq[f] / (mH0_sq[h] - mZ_sq) *(func_I(mf_sq[f],mZ_sq,2) - func_I(mf_sq[f],mH0_sq[h],2));
      result += f_Nc[f] * f_Q[f] * f_coup_Z[2] * temp;
    }
  }
  return result;
}

double C2HDM::Hc_loops() {//B.5
  using exp_c::f_coup_gam;
  using exp_c::f_coup_Z;
  using exp_c::g_Z_Hc_2HDM;
  using exp_c::e; // this is the Hc_gam_gam coupling
  using exp_c::v;
  using exp_c::mZ_sq;

  const double cb = cos(beta);
  const double sb = sin(beta);
  std::array<double,3> g_Hc_h;
  for(size_t i=0; i!=3; i++){
    g_Hc_h[i]  = -(cb*(sb*sb*(Cp.L[2]-Cp.L[5]-Cp.L[6])+cb*cb*Cp.L[4])*Rmix[i][0]
      + sb*(cb*cb*(Cp.L[3]-Cp.L[5]-Cp.L[6])+sb*sb*Cp.L[4])*Rmix[i][1]
      + cb*sb*Cp.L[7]*Rmix[i][2]);
  }
  const std::array<double,3> mH0_sq {mH1*mH1,mH2*mH2,mH3*mH3};
  const double mHc_sq = mHc*mHc;

  double result = 0;
  for(size_t h=0; h!=nHzero; h++){ //iterate over neutral Higgses
    double temp = 0;
    // gammagamma contribution
    temp += f_coup_gam[2] * (-1/2.) * e * v*v / mH0_sq[h] *(
      -func_I(mHc_sq,mH0_sq[h],1)+func_I(mHc_sq,mH0_sq[h],2)
    );
    // Zgamma contribution
    temp += f_coup_Z[2] * (-1/2.) * g_Z_Hc_2HDM * v*v / (mH0_sq[h]- mZ_sq) *(
      func_I(mHc_sq,mZ_sq,1)-func_I(mHc_sq,mH0_sq[h],1)-func_I(mHc_sq,mZ_sq,2)+func_I(mHc*mHc,mH0_sq[h],2)
    );
    temp *= Rhf[h][2][1] * g_Hc_h[h];
    result += temp;
  }
  return result;
}

double C2HDM::W_loops() {//B.7
  using exp_c::mW_sq;
  using exp_c::mZ_sq;
  using exp_c::f_coup_gam;
  using exp_c::f_coup_Z;
  using exp_c::e;
  using exp_c::stw;
  using exp_c::ctw;

  const std::array<double,3> mH0_sq {mH1*mH1,mH2*mH2,mH3*mH3};

  double result = 0;
  for(size_t h=0; h!=nHzero; h++){
    double temp = 0;
    // gammagamma contribution
    temp += f_coup_gam[2] * e * 2 * mW_sq/(mH0_sq[h]) * (
      - 1/4.*( 6 + mH0_sq[h]/mW_sq) * func_I(mW_sq,mH0_sq[h],1)
      + (-4 + 1/4. * (6 + mH0_sq[h]/mW_sq)) * func_I(mW_sq,mH0_sq[h],2)
    );
    //Zgamma contribution
    temp += f_coup_Z[2] * e/stw*ctw * 2 * mW_sq/(mH0_sq[h]-mZ_sq) * (
      -1/4.*((6-mZ_sq/mW_sq)+(1-mZ_sq/(2*mW_sq))*mH0_sq[h]/mW_sq) * (func_I(mW_sq,mH0_sq[h],1)-func_I(mW_sq,mZ_sq,1))
      +((-4+mZ_sq/mW_sq) + 1/4. * (6 - mZ_sq/mW_sq + (1-mZ_sq/(2*mW_sq))*mH0_sq[h]/mW_sq)) * (func_I(mW_sq,mH0_sq[h],2)-func_I(mW_sq,mZ_sq,2))
    );
    result += temp * Rhf[h][2][1] * Rhvv[h];
  }
  return result;
}

double C2HDM::HW_loops() {//B.9
  using exp_c::mW_sq;
  using exp_c::e;
  using exp_c::stw;


  const double mHc_sq = mHc*mHc;
  const std::array<double,3> mH0_sq {mH1*mH1,mH2*mH2,mH3*mH3};
  std::array<double,3> g_Hc_h;
  const double cb = cos(beta);
  const double sb = sin(beta);
  for(size_t i=0; i!=3; i++)
    g_Hc_h[i]  = -(cb*(sb*sb*(Cp[2]-Cp[5]-Cp[6])+cb*cb*Cp[4])*Rmix[i][0]
      + sb*(cb*cb*(Cp[3]-Cp[5]-Cp[6])+sb*sb*Cp[4])*Rmix[i][1]
      + cb*sb*Cp[7]*Rmix[i][2]);

  double result = 0;
  for(size_t h=0; h!=nHzero; h++){
    result += Rhf[h][2][1] * Rhvv[h] * e*e/(2*stw*stw) * mW_sq/(mHc_sq-mW_sq) * (func_I(mW_sq,mH0_sq[h],4) - func_I(mHc_sq,mH0_sq[h],4));
    result += Rhf[h][2][1] * g_Hc_h[h] * mW_sq/(mHc_sq-mW_sq) * (func_I(mW_sq,mH0_sq[h],5)-func_I(mHc_sq,mH0_sq[h],5));
  }
  return result;
}

double C2HDM::func_I(double m1sq, double m2sq, int i) {
  using exp_c::mW_sq;

  double result, error;
  int_params A {m1sq,m2sq,mHc*mHc,mW_sq};

  // Assign gsl function to integrate
  gsl_function F;
  switch(i){
    case 1: F.function = integrand_I1;
        break;
    case 2: F.function = integrand_I2;
        break;
    case 4: F.function = integrand_I4;
        break;
    case 5: F.function = integrand_I5;
        break;
    default:
        clog<<"ScannerS Warning in func_I: called with illegal 3rd argument, returning 0";
        return 0;
        break;
  }
  F.params = &A;
  // perform the integration using an adaptive method
  int status = gsl_integration_qags(&F, 0, 1, 0, 1e-10, 100,EDM_integration, &result, &error);
  //if something went wrong, return inf, which throws the point away (as fabs(inf)>anything)
  if(status!=GSL_SUCCESS){
    clog<<"ScannerS Warning in func_I: GSL integration failure returning inf (dropping the point).\n";
    return 1/0.;
  }
  return result;
}



double integrand_I1(double x, void * params){
  int_params A = *((int_params *) params);
  return (1-2*x*(1-x))*(A.m2sq/(A.m1sq - A.m2sq*x*(1-x))*log(A.m2sq*x*(1-x)/A.m1sq));
}
double integrand_I2(double x, void * params){
  int_params A = *((int_params *) params);
  return A.m2sq/(A.m1sq-A.m2sq*x*(1-x))*log(A.m2sq*x*(1-x)/A.m1sq);
}
double integrand_I4(double x, void * params){
  int_params A = *((int_params *) params);
  return  (x*pow(1-x,2)-4*pow(1-x,2)+(A.mHcsq-A.m2sq)/A.mWsq*x*pow(1-x,2))*(A.m1sq/(A.mWsq*(1-x)+A.m2sq*x-A.m1sq*x*(1-x)))*log((A.mWsq*(1-x)+A.m2sq*x)/(A.m1sq*x*(1-x)));
}
double integrand_I5(double x, void * params){
  int_params A = *((int_params *) params);
  return ( 2*(A.m1sq*x*pow(1-x,2))/(A.mHcsq*(1-x)+A.m2sq*x-A.m1sq*x*(1-x))*log((A.mHcsq*(1-x)+A.m2sq*x)/(A.m1sq*x*(1-x))) );
};
